import datetime
import re
import textwrap
from unittest import mock

import pytest

from hart.exceptions import UserError
from hart.providers import DOProvider, EC2Provider
from hart.roles import get_minion_arguments_for_role, get_provider_for_role, build_minion_id


def test_get_minion_arguments_provider_inheritance(named_tempfile):
    named_tempfile.write(textwrap.dedent('''
        [hart]
        saltmaster = "salt.example.com"
        role_naming_scheme = "{role}.{provider}.example.com"
        salt_branch = "3001"

        [roles.myrole]
        private_networking = true

        [roles.myrole.ec2]
        region = "eu-central-1"
        size = "t3.nano"
        zone = "eu-central-1a"
    ''').encode('utf-8'))
    named_tempfile.close()

    provider = EC2Provider('key_id', 'secret_key')
    arguments = get_minion_arguments_for_role(named_tempfile.name, 'myrole', provider=provider)
    assert is_subdict({
        'minion_id': 'myrole.ec2.example.com',
        'provider': provider,
        'region': 'eu-central-1',
        'size': 't3.nano',
        'private_networking': True,
        'salt_branch': '3001',
        'zone': 'eu-central-1a',
        'minion_config': {
            'master': 'salt.example.com',
            'grains': {
                'roles': [
                    'myrole',
                ],
                'hart.provider': 'ec2',
                'hart.region': 'eu-central-1',
                'hart.size': 't3.nano',
            },
        },
    }, arguments)


def test_get_minion_arguments_region_inheritance(named_tempfile):
    named_tempfile.write(textwrap.dedent('''
        [hart]
        saltmaster = "salt.example.com"
        role_naming_scheme = "{role}.{provider}.example.com"
        salt_branch = "3002"

        [roles.myrole.do]
        size = "s-v2vcpu-2gb"

        [roles.myrole.do.sfo3]
        private_networking = true
        size = "s-4vcpu-4gb"
    ''').encode('utf-8'))
    named_tempfile.close()

    provider = DOProvider('foo')
    arguments = get_minion_arguments_for_role(named_tempfile.name, 'myrole',
        provider=provider, region='sfo3')
    assert is_subdict({
        'minion_id': 'myrole.do.example.com',
        'provider': provider,
        'region': 'sfo3',
        'size': 's-4vcpu-4gb',
        'private_networking': True,
        'salt_branch': '3002',
        'minion_config': {
            'master': 'salt.example.com',
            'grains': {
                'roles': [
                    'myrole',
                ],
                'hart.provider': 'do',
                'hart.region': 'sfo3',
                'hart.size': 's-4vcpu-4gb',
            },
        },
    }, arguments)


def test_get_minion_arguments_for_invalid_role(named_tempfile):
    named_tempfile.write(textwrap.dedent('''
        [roles.myrole]
        private_networking = true
    ''').encode('utf-8'))
    named_tempfile.close()

    with pytest.raises(UserError, match="Unknown role 'foo', must be one of 'myrole'"):
        get_minion_arguments_for_role(named_tempfile.name, 'foo')


def test_get_minion_arguments_for_invalid_role_no_roles(named_tempfile):
    with pytest.raises(UserError, match="Unknown role 'foo', no roles defined in config"):
        get_minion_arguments_for_role(named_tempfile.name, 'foo')


def test_build_templated_minion_id():
    assert build_minion_id('{role}.{zone}.{provider}.example',
        role='foo',
        zone='bar',
        provider='ec2',
    ) == 'foo.bar.ec2.example'


def test_build_templated_minion_id_date_parameters():
    now = datetime.datetime.utcnow()
    assert build_minion_id('{year}{month}{day}.{role}.example',
        role='foo',
    ) == '%s.foo.example' % (now.strftime('%Y%m%d'))


def test_build_unique_minion_id():
    minion_id = build_minion_id('{unique_id}.example')
    assert re.match(r'^[a-f0-9]{8}\.example$', minion_id) is not None


def test_build_minion_id_invalid_parameter():
    with pytest.raises(UserError, match=r"Invalid minion id template variable "
            "{foo}, must be one of {day}, {month}, {role}, {unique_id}, {year}"):
        build_minion_id('{foo}', role='1')


def test_get_minion_arguments_without_provider(named_tempfile):
    named_tempfile.write(textwrap.dedent('''
        [roles.myrole]
        private_networking = true
        provider = "do"
        size = "s-1vcpu-4gb"
        region = "sfo3"

        [providers.do]
        token = "foo"
    ''').encode('utf-8'))
    named_tempfile.close()

    with mock.patch('hart.roles.get_unique_id', lambda: 'unique'):
        arguments = get_minion_arguments_for_role(named_tempfile.name, 'myrole')
    assert arguments['minion_id'] == 'unique.sfo3.do.myrole'
    assert isinstance(arguments['provider'], DOProvider)
    assert arguments['region'] == 'sfo3'
    assert arguments['size'] == 's-1vcpu-4gb'
    assert arguments['private_networking'] == True
    assert is_subdict({
        'grains': {
            'roles': [
                'myrole',
            ],
            'hart.provider': 'do',
            'hart.region': 'sfo3',
            'hart.size': 's-1vcpu-4gb',
        },
    }, arguments['minion_config'])


def test_get_minion_arguments_with_minion_config(named_tempfile):
    named_tempfile.write(textwrap.dedent('''
        [hart.minion_config.grains]
        "environment" = "prod"

        [roles.myrole]
        region = "sfo3"
    ''').encode('utf-8'))
    named_tempfile.close()

    arguments = get_minion_arguments_for_role(named_tempfile.name, 'myrole', provider=DOProvider('foo'))
    assert is_subdict({
        'master_tries': -1,
        'grains': {
            'environment': 'prod',
            'roles': [
                'myrole',
            ],
            'hart.provider': 'do',
            'hart.region': 'sfo3',
            'hart.size': DOProvider.default_size,
        },
    }, arguments['minion_config'])


def test_setting_provider_extensions(named_tempfile):
    named_tempfile.write(textwrap.dedent('''
        [roles.myrole.ec2]
        region = "eu-south-1"
        volume_type = "io1"
    ''').encode('utf-8'))
    named_tempfile.close()

    provider = EC2Provider('key_id', 'secret_key')
    arguments = get_minion_arguments_for_role(named_tempfile.name, 'myrole', provider=provider)

    assert arguments['volume_type'] == 'io1'


def test_use_provider_extension_in_minion_id(named_tempfile):
    named_tempfile.write(textwrap.dedent('''
        [roles.myrole.ec2]
        region = "eu-south-1"
        zone = "eu-south-1a"
        role_naming_scheme = "{zone}.{role}"
    ''').encode('utf-8'))
    named_tempfile.close()

    provider = EC2Provider('key_id', 'secret_key')
    arguments = get_minion_arguments_for_role(named_tempfile.name, 'myrole', provider=provider)

    assert arguments['minion_id'] == 'eu-south-1a.myrole'


def test_use_cli_arguments_in_minion_id(named_tempfile):
    named_tempfile.write(textwrap.dedent('''
        [roles.myrole.ec2]
        region = "eu-south-1"
        role_naming_scheme = "{zone}.{role}"
    ''').encode('utf-8'))
    named_tempfile.close()

    provider = EC2Provider('key_id', 'secret_key')
    arguments = get_minion_arguments_for_role(
        named_tempfile.name, 'myrole', provider=provider, cli_kwargs={
            'zone': 'us-east4-c',
            'provider': 'ec2', # This shouldn't conflict with the manually provided kwarg
        })

    assert arguments['minion_id'] == 'us-east4-c.myrole'


def test_get_provider_from_role(named_tempfile):
    named_tempfile.write(textwrap.dedent('''
        [providers.ec2]
        aws_access_key_id = "foo"
        aws_secret_access_key = "bar"

        [roles.myrole]
        provider = "ec2"
    ''').encode('utf-8'))
    named_tempfile.close()

    provider = get_provider_for_role(named_tempfile.name, 'myrole', None)

    assert isinstance(provider, EC2Provider)


def is_subdict(subset, superset):
    # Kudos to https://stackoverflow.com/a/57675231/5590192 for this, using this
    # for testing to avoid config values added by hart from bloating the test assertions
    if isinstance(subset, dict):
        return all(key in superset and is_subdict(val, superset[key]) for key, val in subset.items())

    # Assume that subset is a plain value if the above doesn't match match
    return subset == superset



from .digitalocean import DOProvider
from .ec2 import EC2Provider
from .vultr import VultrProvider
from .gce import GCEProvider

provider_map = {}
for provider in (
        DOProvider,
        EC2Provider,
        VultrProvider,
        GCEProvider):
    provider_map[provider.alias] = provider



import abc
import contextlib
import time
import json
from collections import namedtuple

import paramiko


NodeSize = namedtuple('NodeSize', 'id memory cpu disk monthly_cost extras')
Region = namedtuple('Region', 'id name')


class BaseProvider(abc.ABC):
    username = 'root'


    def post_connect(self, hart_node):
        pass


    def add_create_minion_arguments(self, parser):
        '''Override this to provide kwargs to create_node'''
        pass


    def add_destroy_minion_arguments(self, parser):
        '''Override this to provide kwargs to destroy_node'''
        pass


    def add_list_regions_arguments(self, parser):
        '''Override this to provide kwargs to get_regions'''
        pass


    def add_list_sizes_arguments(self, parser):
        '''Override this to provide kwargs to get_sizes'''
        pass


    def generate_ssh_key(self): # pylint: disable=no-self-use
        # paramiko doesn't yet support creating Ed25519 keys :'(
        return paramiko.ECDSAKey.generate()


    def wait_for_init_script(self, client, extra=None):
        # Creds to https://stackoverflow.com/a/14158100 for a way to get the pid
        _, stdout, stderr = client.exec_command(
            'echo $$ && exec tail -f -n +1 /var/log/cloud-init-output.log')
        tail_pid = int(stdout.readline())
        for line in stdout:
            print(line, end='')
            if line.startswith('Cloud-init') and ' finished ' in line:
                stdout.channel.close()
                client.exec_command('kill %d' % tail_pid)
                break

        for line in stderr:
            print('Cloud-init stderr: %s' % line.strip())

        _, stdout, stderr = client.exec_command('cat /run/cloud-init/result.json', timeout=3)
        if stdout.channel.recv_exit_status() != 0:
            raise ValueError('Failed to get cloud-init status')

        cloud_init_result = json.loads(''.join(stdout))
        if cloud_init_result['v1']['errors']:
            raise ValueError('cloud-init failed: %s' % ', '.join(cloud_init_result['v1']['errors']))


    def wait_for_public_ip(self, node):
        if node.public_ips and node.public_ips[0] != '0.0.0.0':
            return node
        timeout = 180
        start_time = time.time()
        while True:
            time.sleep(2)
            node = self.get_node(node)
            if node.public_ips and node.public_ips[0] != '0.0.0.0':
                return node
            if time.time() - start_time > timeout:
                raise ValueError('Timed out waiting for node IP: %s' % node.id)


    @contextlib.contextmanager
    def create_temp_ssh_key(self, key_name):
        local_key = self.generate_ssh_key()
        # There's three different variants of the key here, the local key that
        # has the private part, the remote key which has the provider mapping to
        # delete it later, and the auth key, which is passed to the provider
        # again when creating the node to allow authentication.
        public_key = '%s %s' % (local_key.get_name(), local_key.get_base64())
        remote_key, auth_key = self.create_remote_ssh_key(key_name, local_key, public_key)
        print('Created temp ssh key')

        try:
            yield local_key, auth_key
        finally:
            print('Destroying %s ssh key' % self.__class__.__name__)
            self.destroy_remote_ssh_key(remote_key)


    def create_remote_ssh_key(self, key_name, ssh_key, public_key):
        '''Return a tuple of (remote_key, auth_key)'''
        raise NotImplementedError()


    def destroy_remote_ssh_key(self, remote_key):
        raise NotImplementedError()


    def get_size(self, size_name):
        raise NotImplementedError()


    def get_location(self, location_id):
        raise NotImplementedError()


    def destroy_node(self, node, extra=None, **kwargs):
        raise NotImplementedError()


    def get_node(self, node):
        '''node can be either a id or a Node as returned from `create_node`.'''
        raise NotImplementedError()


    def get_regions(self, **kwargs):
        raise NotImplementedError()


    def get_sizes(self, **kwargs):
        raise NotImplementedError()


    def create_node(self,
            minion_id,
            region,
            debian_codename,
            auth_key,
            cloud_init,
            private_networking,
            tags,
            size=None,
            **kwargs):
        raise NotImplementedError()



import tempfile
import textwrap

from hart.providers import DOProvider
from hart.config import build_provider_from_config, build_provider_from_file


def test_build_provider_from_file(named_tempfile):
    named_tempfile.write(textwrap.dedent('''
        [providers.do]
        token = "foo"
    ''').encode('utf-8'))
    named_tempfile.close()

    provider = build_provider_from_file('do', named_tempfile.name)
    assert isinstance(provider, DOProvider)


def test_build_provider_from_config():
    provider = build_provider_from_config('do', {
        'providers': {
            'do': {'token': 'foo'},
        },
    })
    assert isinstance(provider, DOProvider)



from unittest import mock

import pytest

from hart.providers import vultr


def test_get_device_from_interfaces():
    device_name, next_label = vultr.get_device_and_next_label_from_interfaces({
        'lo': {
            'inet': [{
                'address': '127.0.0.1',
                'label': 'lo',
            }]
        },
        'ens3': {
            'inet': [{
                'address': '1.2.3.4',
            }, {
                'address': '2.3.4.5',
                'label': 'ens3:0',
            }]
        }
    }, '1.2.3.4')
    assert device_name == 'ens3'
    assert next_label == 'ens3:1'


def test_add_private_ip_to_device():
    with mock.patch('hart.providers.vultr.subprocess') as mock_subprocess:
        vultr.add_ip_to_device('minion', 'private', 'ens7', '10.0.0.1', '255.255.240.0')
    mock_subprocess.run.assert_called_with([
        'salt',
        'minion',
        'file.write',
        '/etc/network/interfaces.d/20-hart-private-ip',
        "args=['auto ens7', 'iface ens7 inet static', 'address 10.0.0.1', 'netmask 255.255.240.0', 'mtu 1450']",
    ], check=True)


def test_add_reserved_ip_to_device():
    with mock.patch('hart.providers.vultr.subprocess') as mock_subprocess:
        vultr.add_ip_to_device('minion', 'reserved', 'ens3:0', '1.2.3.4', '255.255.255.0')
    mock_subprocess.run.assert_called_with([
        'salt',
        'minion',
        'file.write',
        '/etc/network/interfaces.d/20-hart-reserved-ip',
        "args=['auto ens3:0', 'iface ens3:0 inet static', 'address 1.2.3.4', 'netmask 255.255.255.0']",
    ], check=True)


def test_get_device_from_missing_interface():
    with pytest.raises(ValueError):
        vultr.get_device_and_next_label_from_interfaces({}, '1.2.3.4')



import argparse

from hart import utils as uut

def test_remove_argument_from_parser_single_flag():
    parser = argparse.ArgumentParser()
    parser.add_argument('--foo')
    assert '--foo' in parser.format_usage()

    uut.remove_argument_from_parser(parser, '--foo')

    assert '--foo' not in parser.format_usage()

    # This shouldn't crash
    parser.add_argument('--foo', help='Something')


def test_remove_argument_from_parser_multiple_flags():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--foo')
    assert '-f' in parser.format_usage()

    uut.remove_argument_from_parser(parser, '--foo')

    assert '-f' not in parser.format_usage()
    assert '--foo' not in parser.format_usage()

    # This shouldn't crash
    parser.add_argument('-f', '--foo', help='Something')



from unittest.mock import Mock

import pytest

from hart.exceptions import UserError
from hart.providers import gce


def test_get_selected_or_default_subnet_none():
    with pytest.raises(UserError):
        gce.get_selected_or_default_subnet([], 'irrelevant')


def test_get_only_alternative():
    first = Mock()
    first.name = 'first'
    ret = gce.get_selected_or_default_subnet([first], None)
    assert ret.name == 'first'


def test_get_several_alternatives():
    first = Mock()
    first.name = 'first'
    second = Mock()
    second.name = 'second'
    with pytest.raises(UserError):
        gce.get_selected_or_default_subnet([first, second], None)


def test_get_selected_or_default_subnet_multiple():
    first = Mock()
    first.name = 'first'
    second = Mock()
    second.name = 'second'
    ret = gce.get_selected_or_default_subnet([first, second], 'first')
    assert ret.name == 'first'


def test_get_selected_or_default_subnet_none_matching():
    first = Mock()
    first.name = 'first'
    second = Mock()
    second.name = 'second'
    with pytest.raises(UserError):
        gce.get_selected_or_default_subnet([first, second], 'third')


def test_instance_name_generation():
    minion_id = '01.db.example.com'
    assert gce.name_from_minion_id(minion_id) == 'hart-com-example-db-01-b64faa'



import binascii
import datetime
import os

from .config import load_config, build_provider_from_config
from .exceptions import UserError

DEFAULT_MINION_NAMING_SCHEME = '{unique_id}.{region}.{provider}.{role}'

def get_provider_for_role(config_file, role, region):
    config = load_config(config_file)
    core_config = config.get('hart', {})
    role_config = get_role_config(config, role)

    merged_config = {}
    merged_config.update(core_config)

    if region is None:
        region = merged_config.pop('region', None)

    provider_alias = merged_config.pop('provider', role_config.pop('provider', None))
    return build_provider_from_config(provider_alias, config, region=region)


def get_minion_arguments_for_role(config_file, role, provider=None, region=None, cli_kwargs=None):
    if cli_kwargs is None:
        cli_kwargs = {}

    config = load_config(config_file)
    core_config = config.get('hart', {})
    role_config = get_role_config(config, role)

    merged_config = {}
    merged_config.update(core_config)

    if region is None:
        region = merged_config.pop('region', None)

    provider_alias = merged_config.pop('provider', role_config.pop('provider', None))
    if provider is None:
        provider = build_provider_from_config(provider_alias, config, region=region)

    provider_config = role_config.pop(provider.alias, {})
    merged_config.update(role_config)

    region_config = provider_config.pop(region, {})
    merged_config.update(provider_config)
    merged_config.update(region_config)
    merged_config.update(cli_kwargs)

    # We might not know the region until getting the provider config, thus try
    # to get it again
    if region is None:
        region = merged_config.pop('region')
    else:
        merged_config.pop('region', None)

    size = merged_config.setdefault('size', provider.default_size)

    default_minion_config = {
        'grains': {
            'roles': [role],
            'hart.region': region,
            'hart.provider': provider.alias,
            'hart.size': size,
        },
        # Default to keep retrying a master connection if it fails
        'master_tries': -1,
        # This config is necessary to make salt try to reconnect to a master on
        # a new IP in case of failover
        'master_alive_interval': 90,
        'auth_safemode': False,
        'master_type': 'failover',
        'retry_dns': 0,
    }

    saltmaster = merged_config.pop('saltmaster', None)
    if saltmaster:
        default_minion_config['master'] = saltmaster

    minion_config = merged_config.pop('minion_config', {})
    if minion_config:
        merge_dicts(default_minion_config, minion_config)

    naming_scheme = merged_config.pop('role_naming_scheme', DEFAULT_MINION_NAMING_SCHEME)

    # Prevent duplicate provider error, added back again soon
    merged_config.pop('provider', None)

    merged_config['minion_id'] = build_minion_id(naming_scheme,
        role=role,
        region=region,
        provider=provider.alias,
        **merged_config,
    )
    merged_config['provider'] = provider
    merged_config['region'] = region
    merged_config['minion_config'] = default_minion_config

    return merged_config


def get_role_config(config, role):
    roles = config.get('roles', {})
    role_config = roles.get(role)
    if role_config is None:
        if roles:
            raise UserError('Unknown role %r, must be one of %s' % (
                role, ', '.join(repr(r) for r in config.get('roles', {}))))
        else:
            raise UserError('Unknown role %r, no roles defined in config' % role)
    return role_config


def build_minion_id(naming_scheme, **kwargs):
    now = datetime.datetime.utcnow()
    kwargs['unique_id'] = get_unique_id()
    kwargs['year'] = now.strftime('%Y')
    kwargs['month'] = now.strftime('%m')
    kwargs['day'] = now.strftime('%d')
    try:
        return naming_scheme.format(**kwargs)
    except KeyError as error:
        raise UserError('Invalid minion id template variable {%s}, must be one of %s' % (
            error.args[0], ', '.join('{%s}' % s for s in sorted(kwargs))))


def merge_dicts(a, b):
    # b overwrites leaf values in a
    for key, val in b.items():
        if isinstance(val, dict):
            a[key] = merge_dicts(a[key], val)
        else:
            a[key] = val
    return a


def get_unique_id():
    return binascii.hexlify(os.urandom(4)).decode('utf-8')



#!./venv/bin/python

import json
import os
import subprocess
import sys
import time
import traceback

import yaml

from . import utils
from .constants import DEBIAN_VERSIONS
from .ssh import get_verified_ssh_client, ssh_run_command
from .utils import log_error


def create_minion(
        minion_id,
        provider,
        region=None,
        size=None,
        salt_branch='latest',
        debian_codename='bullseye',
        tags=None,
        private_networking=False,
        minion_config=None,
        **kwargs
        ):
    hart_node = create_node(
        minion_id,
        provider,
        region,
        size,
        salt_branch,
        debian_codename,
        tags,
        private_networking,
        minion_config,
        **kwargs
    )
    try:
        connect_minion(hart_node)
    except:
        log_error('Destroying node since it failed to connect')
        hart_node.provider.destroy_node(hart_node.node, extra=hart_node.node_extra)
        disconnect_minion(minion_id)
        raise


def connect_minion(hart_node):
    username = hart_node.provider.username
    with get_verified_ssh_client(
            hart_node.public_ip,
            hart_node.ssh_key,
            hart_node.ssh_canary,
            username) as client:
        hart_node.provider.wait_for_init_script(client, hart_node.node_extra)
        minion_pubkey = get_minion_pubkey(client, should_sudo=username != 'root')
        trust_minion_key(hart_node.minion_id, minion_pubkey)
        print('Minion added: %s' % hart_node.public_ip)
        verify_minion_connection(client, hart_node.minion_id, username)
        hart_node.provider.post_connect(hart_node)


def create_node(
        minion_id,
        provider,
        region=None,
        size=None,
        salt_branch='latest',
        debian_codename='bullseye',
        tags=None,
        private_networking=False,
        minion_config=None,
        **kwargs
        ):
    ssh_canary = utils.create_token()
    cloud_init_template = utils.get_cloud_init_template()
    master_pubkey = get_master_pubkey()
    default_minion_config = {
        'id': minion_id,
    }
    if minion_config is not None:
        default_minion_config.update(minion_config)

    saltstack_repo = utils.get_saltstack_repo_url(debian_codename, salt_branch)
    cloud_init = cloud_init_template.render(**{
        'random_seed': utils.create_token(),
        'minion_config': yaml.dump(default_minion_config),
        'ssh_canary': ssh_canary,
        'master_pubkey': master_pubkey,
        'saltstack_repo': saltstack_repo,
        'wait_for_apt': DEBIAN_VERSIONS[debian_codename] >= 10,
        'permit_root_ssh': provider.username == 'root',
    })

    key_name = utils.build_ssh_key_name(minion_id)

    if not check_existing_minion(minion_id):
        print('Existing minions were found and did want to overwrite, aborting')
        return

    with provider.create_temp_ssh_key(key_name) as (ssh_key, auth_key):
        node = None
        if size:
            kwargs['size'] = size
        try:
            node, extra = provider.create_node(
                minion_id,
                region,
                debian_codename,
                auth_key,
                cloud_init,
                private_networking,
                tags,
                **kwargs)
            node = provider.wait_for_public_ip(node)
            public_ip = node.public_ips[0]
            print('Node running at %s' % public_ip)
            return utils.HartNode(minion_id, public_ip, node, provider, ssh_key, ssh_canary, extra)
        except:
            traceback.print_exc()
            if node:
                log_error('Destroying node since it failed initialization')
                provider.destroy_node(node, extra)
            raise


def destroy_minion(minion_id, provider, **kwargs):
    disconnect_minion(minion_id)
    print('Destroying minion')
    node = provider.get_node(minion_id)
    provider.destroy_node(node, **kwargs)


def disconnect_minion(minion_id):
    print('Deleting the salt minion %s' % minion_id)
    subprocess.run([
        'salt-key',
        '--delete=%s' % minion_id,
        '--yes',
    ], check=True)


def destroy_node(hart_node):
    hart_node.provider.destroy_node(hart_node.node, extra=hart_node.node_extra)


def get_master_pubkey():
    with open('/etc/salt/pki/master/master.pub') as fh:
        return fh.read()


def check_existing_minion(minion_id):
    minions = json.loads(subprocess.check_output([
        'salt-key',
        '--print', minion_id,
        '--out=json',
    ]).decode('utf-8'))

    existing_categories = minions.keys()
    if existing_categories:
        should_continue = input('Existing minions matching %s were found in %s, '
            'overwrite? [y/N]' % (minion_id, ', '.join(existing_categories)))
        return should_continue == 'y'

    return True


def trust_minion_key(minion_id, minion_pubkey):
    with open('/etc/salt/pki/master/minions/%s' % minion_id, 'wb') as fh:
        fh.write(minion_pubkey.encode('utf-8'))
        os.fchmod(fh.fileno(), 0o644)

    # If the minion connected before we trusted the key, remove the duplicate key in minions_pre
    pre_key_path = '/etc/salt/pki/master/minions_pre/%s' % minion_id
    try:
        pre_fh = open(pre_key_path)
    except IOError:
        return

    with pre_fh:
        pre_key = pre_fh.read()

    if pre_key.strip() == minion_pubkey.strip():
        print('Got early connection attempt from minion, removing pre-accept key')
        os.remove(pre_key_path)


def verify_minion_connection(client, minion_id, username):
    # The restart is needed since the minion might have attempted connecting to
    # the salt master before the key got trusted and thus might not be ready to
    # respond to a ping from the master
    prefix = 'sudo ' if username != 'root' else ''
    ssh_run_command(client,
        '{0}salt-call test.ping && {0}service salt-minion restart'.format(prefix),
        timeout=120)

    # Also test that the master can reach the minion, but the minion might take a moment
    # to start so try a couple times
    for i in range(5):
        try:
            subprocess.run(['salt', minion_id, 'test.ping'],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as e:
            if e.returncode == 1 and (b'[Not connected]' in e.stdout or b'[No response]' in e.stdout):
                print('Failed connectivity check %d, trying again...' % (i + 1))
                time.sleep(2**i)
                continue
            print(e.stdout.decode('utf-8'))
            log_error(e.stderr.decode('utf-8'))
            raise
        else:
            print('Pinged %s successfully' % minion_id)
            break
    else:
        raise ConnectionError('Unable to ping new instance')

    authorized_keys_path = '/root/.ssh/authorized_keys'
    if username != 'root':
        authorized_keys_path = '/home/%s/.ssh/authorized_keys' % username
    ssh_run_command(client, 'rm %s' % authorized_keys_path)


def get_minion_pubkey(client, should_sudo):
    cmd = '%scat /etc/salt/pki/minion/minion.pub' % ('sudo ' if should_sudo else '')
    return ssh_run_command(client, cmd, log_stdout=False)



import sys
import traceback

import yaml

from . import utils
from .constants import DEBIAN_VERSIONS
from .ssh import get_verified_ssh_client, ssh_run_command


def create_master(
        minion_id,
        provider,
        region=None,
        size=None,
        salt_branch='latest',
        debian_codename='bookworm',
        tags=None,
        private_networking=False,
        minion_config=None,
        grains=None,
        script=None,
        authorize_key=None,
        **kwargs
        ):
    hart_node = create_master_node(
        minion_id,
        provider,
        region,
        size,
        salt_branch,
        debian_codename,
        tags,
        private_networking,
        minion_config,
        grains,
        **kwargs
    )
    try:
        connect_to_master(hart_node, script, authorize_key)
    except:
        sys.stderr.write('Destroying master since it failed startup\n')
        hart_node.provider.destroy_node(hart_node.node, extra=hart_node.node_extra)
        raise


def create_master_node(
        minion_id,
        provider,
        region=None,
        size=None,
        salt_branch='latest',
        debian_codename='bookworm',
        tags=None,
        private_networking=False,
        minion_config=None,
        grains=None,
        **kwargs
        ):
    ssh_canary = utils.create_token()
    cloud_init_template = utils.get_cloud_init_template('master.sh')
    default_minion_config = {
        'id': minion_id,
        'user': 'salt',
        'file_client': 'local',
        'state_verbose': False,
    }
    if minion_config is not None:
        default_minion_config.update(minion_config)

    saltstack_repo = utils.get_saltstack_repo_url(debian_codename, salt_branch)
    cloud_init = cloud_init_template.render(**{
        'random_seed': utils.create_token(),
        'minion_config': yaml.dump(default_minion_config),
        'grains': yaml.dump({'grains': grains}) if grains else None,
        'ssh_canary': ssh_canary,
        'saltstack_repo': saltstack_repo,
        'wait_for_apt': DEBIAN_VERSIONS[debian_codename] >= 10,
        'permit_root_ssh': provider.username == 'root',
        'add_user': salt_branch != 'latest' and int(salt_branch[:salt_branch.find('.')]) < 3006,
    })

    key_name = utils.build_ssh_key_name(minion_id)

    with provider.create_temp_ssh_key(key_name) as (ssh_key, auth_key):
        node = None
        if size:
            kwargs['size'] = size
        try:
            node, extra = provider.create_node(
                minion_id,
                region,
                debian_codename,
                auth_key,
                cloud_init,
                private_networking,
                tags,
                **kwargs)
            node = provider.wait_for_public_ip(node)
            public_ip = node.public_ips[0]
            print('Master running at %s' % public_ip)
            return utils.HartNode(minion_id, public_ip, node, provider, ssh_key, ssh_canary, extra)
        except:
            traceback.print_exc()
            if node:
                sys.stderr.write('Destroying node since it failed initialization\n')
                provider.destroy_node(node, extra)
            raise


def connect_to_master(hart_node, script, authorize_key=None):
    username = hart_node.provider.username
    with get_verified_ssh_client(
            hart_node.public_ip,
            hart_node.ssh_key,
            hart_node.ssh_canary,
            username) as client:
        hart_node.provider.wait_for_init_script(client, hart_node.node_extra)
        if authorize_key:
            ssh_run_command(client, 'echo "%s" >> ~/.ssh/authorized_keys' % authorize_key)
        if script:
            sftp_client = client.open_sftp()
            # Preserving this on disk as a record of how the master was created
            script_path = '/root/hart-master-init'
            with sftp_client.file(script_path, 'wx') as remote_file:
                with open(script, 'rb') as local_file:
                    chunk_size = 16*2**10
                    for chunk in iter(lambda: local_file.read(chunk_size), b''):
                        remote_file.write(chunk)
            sftp_client.chmod(script_path, 0o700)
            sftp_client.close()
            ssh_run_command(client, script_path, timeout=None)

        master_pubkeys = ssh_run_command(client,
            'for pubkey in /etc/ssh/ssh_host_*_key.pub; do ssh-keygen -lf "$pubkey"; done',
            log_stdout=False)
        print('Master created: %s@%s\nssh fingerprints: \n%s' % (
            username, hart_node.public_ip, master_pubkeys))



import ipaddress
import json
import datetime
import sys
import time

import boto3
import ifaddr
import paramiko
from libcloud.compute.base import Node

from .base import BaseProvider, NodeSize, Region
from ..constants import DEBIAN_VERSIONS
from ..exceptions import UserError
from ..utils import remove_argument_from_parser


# The pricing API is a supreme clusterfuck that requires lots of special care.
# One particular issue we run into when listing EC2 instance sizes in a given
# region is that the API doesn't allow you to filter on region, but on
# "location", which is the human-readable name of the region. There's no API to
# get the mapping between regions and the name, thus you either have to hardcode
# it like we do here and be forced to push and update whenever Amazon adds
# another region, or scrape the
# https://docs.aws.amazon.com/general/latest/gr/rande.html#ec2_region page,
# which will break if they change the html structure of that page. Opted for the
# hardcoding here for simplicity.
region_to_location_map = {
    'af-south-1': 'Africa (Cape Town)',
    'ap-east-1': 'Asia Pacific (Hong Kong)',
    'ap-northeast-1': 'Asia Pacific (Tokyo)',
    'ap-northeast-2': 'Asia Pacific (Seoul)',
    'ap-northeast-3': 'Asia Pacific (Osaka)',
    'ap-south-1': 'Asia Pacific (Mumbai)',
    'ap-southeast-1': 'Asia Pacific (Singapore)',
    'ap-southeast-2': 'Asia Pacific (Sydney)',
    'ca-central-1': 'Canada (Central)',
    'cn-north-1': 'China (Beijing)',
    'cn-northwest-1': 'China (Ningxia)',
    'eu-central-1': 'EU (Frankfurt)',
    'eu-north-1': 'EU (Stockholm)',
    'eu-south-1': 'EU (Milan)',
    'eu-west-1': 'EU (Ireland)',
    'eu-west-2': 'EU (London)',
    'eu-west-3': 'EU (Paris)',
    'me-south-1': 'Middle East (Bahrain)',
    'sa-east-1': 'South America (Sao Paulo)',
    'us-east-1': 'US East (N. Virginia)',
    'us-east-2': 'US East (Ohio)',
    'us-west-1': 'US West (N. California)',
    'us-west-2': 'US West (Oregon)',
}


class EC2Provider(BaseProvider):
    username = 'admin'
    alias = 'ec2'
    default_size = 't3.micro'

    def __init__(self, aws_access_key_id, aws_secret_access_key, region=None):
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.region = region
        self._ec2 = None


    @property
    def ec2(self):
        # Construct this lazily to prevent the region from having to be
        # specified to list regions or sizes
        if self._ec2:
            return self._ec2

        self._ec2 = boto3.client('ec2',
            region_name=self.region,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )
        return self._ec2


    def generate_ssh_key(self):
        # EC2 only supports RSA for ssh keys
        return paramiko.RSAKey.generate(2048)


    def add_create_minion_arguments(self, parser):
        def split_csv_keyval(clistring):
            ret = {}
            for key_value_pair in clistring.split(','):
                if not '=' in key_value_pair:
                    raise UserError('EC2 requires tags to be key=value pairs')

                key, value = key_value_pair.split('=', 1)
                ret[key] = value
            return ret

        remove_argument_from_parser(parser, '--tags')
        parser.add_argument('-t', '--tags', type=split_csv_keyval, default={},
            help='Tags to add to the new node, comma-separated list of key=value pairs.')

        parser.add_argument('-z', '--zone', help='AWS availability zone')
        parser.add_argument('--subnet',
            help='AWS: The subnet to launch the node in')
        parser.add_argument('--volume-size', type=int,
            help='The size of the EBS drive')
        parser.add_argument('--volume-type',
            choices=('standard', 'io1', 'gp2', 'sc1', 'st1'),
            help='The type of EBS drive to mount')
        parser.add_argument('--volume-iops', type=int,
            help='How many IOPS to provision (only applies to io1 volumes)')
        parser.add_argument('--connection-gateway',
            help="If the saltmaster's outbound IP can't be automatically detected, "
            "specify the CIDR range to allow through the firewall to the minion here.")


    def add_list_regions_arguments(self, parser):
        parser.add_argument('-z', '--include-zones', action='store_true')


    def create_remote_ssh_key(self, key_name, ssh_key, public_key):
        import_response = self.ec2.import_key_pair(KeyName=key_name, PublicKeyMaterial=public_key)
        assert import_response['ResponseMetadata']['HTTPStatusCode'] == 200
        return key_name, key_name


    def destroy_remote_ssh_key(self, remote_key):
        self.ec2.delete_key_pair(KeyName=remote_key)


    def create_node(self,
            minion_id,
            region,
            debian_codename,
            auth_key,
            cloud_init,
            private_networking,
            tags,
            size=None,
            **kwargs):
        if size is None:
            size = self.default_size

        zone = kwargs.get('zone')
        if not zone:
            raise UserError('You must specify the ec2 availability zone')

        size = self.get_size(size)
        image = self.get_image(debian_codename)

        subnet = kwargs.get('subnet')
        subnet_ids = [subnet] if subnet else []
        subnet_response = self.ec2.describe_subnets(SubnetIds=subnet_ids,
            Filters=[{'Name': 'availability-zone', 'Values': [zone]}])
        subnets = subnet_response['Subnets']

        if not subnets and subnet:
            raise UserError('No subnet matching %s in %s' % (subnet, zone))
        elif not subnets:
            raise UserError('No available subnets in %s' % zone)
        elif len(subnets) > 1:
            subnet_summary = []
            for subnet in subnets:
                for tag in subnet.get('Tags', []):
                    if tag['Key'] == 'Name':
                        subnet_summary.append('%s (%s)' % (subnet['SubnetId'], tag['Value']))
                        break
                else:
                    subnet_summary.append(subnet['SubnetId'])
            raise UserError('More than one subnet in availability zone, specify'
                ' which one to use: %s' % (', '.join(subnet_summary)))

        subnet = subnets[0]
        temp_security_group = self.create_temp_security_group(minion_id,
            kwargs.get('connection_gateway'), subnet['VpcId'])

        block_devices = []
        volume_type = kwargs.get('volume_type')
        volume_size = kwargs.get('volume_size')
        if volume_type or volume_size:
            ebs = {}
            # iops is only supported for io1 volumes and thus can only be
            # specified with --volume-type
            iops = kwargs.get('volume_iops')
            if iops:
                ebs['Iops'] = iops

            if volume_type:
                ebs['VolumeType'] = volume_type

            # The debian images automatically expand the filesystem to match on boot
            if volume_size:
                ebs['VolumeSize'] = volume_size
            block_devices.append({
                'DeviceName': image['RootDeviceName'],
                'Ebs': ebs,
            })

        tag_specifications = [{'Key': 'Name', 'Value': minion_id}]
        for key, val in tags.items():
            tag_specifications.append({
                'Key': key,
                'Value': val,
            })

        create_response = self.ec2.run_instances(
            ImageId=image['ImageId'],
            InstanceType=size,
            KeyName=auth_key,
            Placement={'AvailabilityZone': zone},
            UserData=cloud_init,
            MinCount=1,
            MaxCount=1,
            BlockDeviceMappings=block_devices,
            NetworkInterfaces=[{
                'AssociatePublicIpAddress': True,
                'DeleteOnTermination': True,
                'DeviceIndex': 0,
                'SubnetId': subnet['SubnetId'],
                'Groups': [temp_security_group],
            }],
            TagSpecifications=[{
                'ResourceType': 'instance',
                'Tags': tag_specifications,
            }],
        )
        instance = create_response['Instances'][0]

        node = Node(id=instance['InstanceId'], name=minion_id, state=instance['State']['Name'],
            public_ips=[], private_ips=[instance['PrivateIpAddress']],
            driver=self.ec2, created_at=instance['LaunchTime'], extra=None)
        return node, {
            'groupId': temp_security_group,
            'vpcId': subnet['VpcId'],
        }


    def get_node(self, node):
        if isinstance(node, str):
            instance_response = self.ec2.describe_instances(Filters=[{
                'Name': 'tag:Name',
                'Values': [node],
            }])
        else:
            # This can fail if called right after run_instances, retry if not found
            start_time = time.time()
            exception = None
            while time.time() - start_time < 20:
                try:
                    instance_response = self.ec2.describe_instances(InstanceIds=[node.id])
                    break
                except Exception as error:
                    exception = error
                    time.sleep(1)
            else:
                # no break
                raise ValueError('Instance not found') from exception

        instance = instance_response['Reservations'][0]['Instances'][0]
        public_ip = instance.get('PublicIpAddress')
        public_ips = [public_ip] if public_ip else []
        name = None
        for tag in instance['Tags']:
            if tag['Key'] == 'Name':
                name = tag['Value']
                break
        private_ip = instance.get('PrivateIpAddress')
        private_ips = [private_ip] if private_ip else []
        return Node(id=instance['InstanceId'], name=name, state=instance['State']['Name'],
            public_ips=public_ips, private_ips=private_ips, driver=self.ec2,
            created_at=instance['LaunchTime'], extra=None)


    def create_temp_security_group(self, minion_id, connection_gateway, vpc_id):
        current_date = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H-%M-%S')
        name = 'temp-for-%s-%s' % (minion_id, current_date)

        # Get the external IPs for the current host to let through the firewall
        # to the minion for the initial ssh connection
        if connection_gateway:
            external_ips = [connection_gateway]
        else:
            external_ips = list(get_host_public_ips())

        if not external_ips:
            raise UserError('Could not find any public IPs on the current '
                'host and thus wont be able to connect to the new node')

        group = self.ec2.create_security_group(
            GroupName=name,
            Description='Temporary group for initial saltmaster ssh initialization',
            VpcId=vpc_id,
        )
        group_id = group['GroupId']

        self.ec2.authorize_security_group_ingress(
            GroupId=group_id,
            IpPermissions=[{
                'IpProtocol': 'tcp',
                'FromPort': 22,
                'ToPort': 22,
                'IpRanges': [{'CidrIp': ip} for ip in external_ips],
            }]
        )

        return group_id


    def get_image(self, debian_codename):
        debian_version = DEBIAN_VERSIONS[debian_codename]
        official_debian_account = '136693071363'
        image_response = self.ec2.describe_images(Owners=[official_debian_account], Filters=[{
            'Name': 'architecture',
            'Values': ['x86_64'],
        }])
        image_prefix = 'debian-%s' % debian_version
        dist_images = [i for i in image_response['Images'] if i['Name'].startswith(image_prefix)]
        dist_images.sort(key=lambda i: i['Name'])
        return dist_images[-1]


    def post_connect(self, hart_node):
        # Delete the temp security group that allowed ssh
        # Detach the security group from the instance. An instance must have at
        # least one security group, so we attach the VPC default group.
        self.delete_node_security_group(
            hart_node.node,
            hart_node.node_extra['groupId'],
            hart_node.node_extra['vpcId'],
        )


    def delete_node_security_group(self, node, group_id, vpc_id):
        response = self.ec2.describe_security_groups(
            Filters=[{
                'Name': 'group-name',
                'Values': ['default'],
            }, {
                'Name': 'vpc-id',
                'Values': [vpc_id],
            }],
        )
        default_group = response['SecurityGroups'][0]
        self.ec2.modify_instance_attribute(InstanceId=node.id, Groups=[default_group['GroupId']])
        print('Deleting temporary security group')
        response = self.ec2.delete_security_group(GroupId=group_id)


    def destroy_node(self, node, extra=None, **kwargs):
        if extra is not None:
            self.delete_node_security_group(node, extra['groupId'], extra['vpcId'])

        self.ec2.terminate_instances(InstanceIds=[node.id])


    def get_size(self, size_name):
        return size_name


    def get_sizes(self, **kwargs):
        region = self.region
        if region is None:
            sys.stderr.write('No region specified, using us-east-1. Sizes listed might not exist '
                'in any other region\n')
            region = 'us-east-1'

        pricing = boto3.client('pricing',
            region_name='us-east-1',
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )
        sizes = []
        location = region_to_location_map[region]
        filters = [
            {'Field': 'currentGeneration', 'Value': 'Yes', 'Type': 'TERM_MATCH'},
            {'Field': 'operatingSystem', 'Value': 'Linux', 'Type': 'TERM_MATCH'},
            {'Field': 'location', 'Value': location, 'Type': 'TERM_MATCH'},
            {'Field': 'productFamily', 'Value': 'Compute Instance', 'Type': 'TERM_MATCH'},
            {'Field': 'capacityStatus', 'Value': 'Used', 'Type': 'TERM_MATCH'},
            {'Field': 'preInstalledSw', 'Value': 'NA', 'Type': 'TERM_MATCH'},
            {'Field': 'tenancy', 'Value': 'Shared', 'Type': 'TERM_MATCH'},
        ]
        paginator = pricing.get_paginator('get_products')
        results = paginator.paginate(
                ServiceCode='AmazonEC2',
                Filters=filters,
                FormatVersion='aws_v1',
        )
        for response in results:
            for stringified_data in response['PriceList']:
                data = json.loads(stringified_data)
                attributes = data['product']['attributes']
                extras = {
                    'family': attributes['instanceFamily'],
                    'network': attributes['networkPerformance'],
                }
                if 'cpuFreq' in attributes:
                    extras['cpuFreq'] = attributes['clockSpeed']

                offer = data['terms']['OnDemand'].popitem()[1]
                price_dimension = offer['priceDimensions'].popitem()[1]
                price_per_unit = float(price_dimension['pricePerUnit']['USD'])
                price_unit = price_dimension['unit']
                if price_unit == 'Hrs':
                    price = price_per_unit * 720
                else:
                    raise ValueError('Unknown price unit: %s' % price_unit)

                memory_value, memory_unit = attributes['memory'].split()
                if memory_unit == 'GiB':
                    memory = float(memory_value.replace(',', ''))
                else:
                    raise ValueError('unknown memory unit: %s' % memory_unit)

                sizes.append(NodeSize(
                    attributes['instanceType'],
                    memory,
                    int(attributes['vcpu']),
                    attributes['storage'],
                    price,
                    extras,
                ))

        sizes.sort(key=lambda s: s.monthly_cost)
        return sizes


    def get_regions(self, include_zones=False, **kwargs):
        regions = []
        # If not region is specified, pick an arbitrary one that is probably
        # active (ie added before March 20, 2019)
        if not self.region:
            self.region = 'us-west-1'

        response = self.ec2.describe_regions()
        for region in response['Regions']:
            if include_zones:
                region_boto = boto3.client('ec2',
                    region_name=region['RegionName'],
                    aws_access_key_id=self.aws_access_key_id,
                    aws_secret_access_key=self.aws_secret_access_key,
                )
                az_response = region_boto.describe_availability_zones()
                for zone in az_response['AvailabilityZones']:
                    # Don't fail if we don't know the name of the region to avoid
                    # crashing for new regions we don't know about yet
                    region_name = region_to_location_map.get(zone['RegionName'], 'Unknown')
                    regions.append(Region(zone['ZoneName'], region_name))
            else:
                region_id = region['RegionName']
                name = region_to_location_map.get(region_id, 'Unknown')
                regions.append(Region(region_id, name))
        regions.sort(key=lambda r: r.name)
        return regions


def get_host_public_ips():
    for adapter in ifaddr.get_adapters():
        for ip in adapter.ips:
            string_ip = ip.ip[0] if isinstance(ip.ip, tuple) else ip.ip
            address = ipaddress.ip_address(string_ip)
            if address.is_global:
                yield '%s/%s' % (string_ip, address.max_prefixlen)



import hashlib

from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider

from .base import NodeSize, Region
from .libcloud import BaseLibcloudProvider
from ..constants import DEBIAN_VERSIONS
from ..exceptions import UserError

# Haven't found a way to get pretty location names from the API yet, thus
# hardcoding these where we know them
region_pretty_names = {
    'asia-east1': 'Changhua County, Taiwan',
    'asia-east2': 'Hong Kong',
    'asia-northeast1': 'Tokyo, Japan',
    'asia-northeast2': 'Osaka, Japan',
    'asia-northeast3': 'Seoul, South Korea',
    'asia-south1': 'Mumbai, India',
    'asia-south2': 'Delhi, India',
    'asia-southeast1': 'Jurong West, Singapore',
    'asia-southeast2': 'Jakarta, Indonesia',
    'australia-southeast1': 'Sydney, Australia',
    'australia-southeast2': 'Melbourne, Australia',
    'europe-central2': 'Warsaw, Poland',
    'europe-north1': 'Hamina, Finland',
    'europe-west1': 'St. Ghislain, Belgium',
    'europe-west2': 'London, England, UK',
    'europe-west3': 'Frankfurt, Germany',
    'europe-west4': 'Eemshaven, Netherlands',
    'europe-west6': 'Zürich, Switzerland',
    'northamerica-northeast1': 'Montréal, Québec, Canada',
    'northamerica-northeast2': 'Toronto, Ontario',
    'southamerica-east1': 'Osasco (São Paulo), Brazil',
    'us-central1': 'Council Bluffs, Iowa, USA',
    'us-east1': 'Moncks Corner, South Carolina, USA',
    'us-east4': 'Ashburn, Northern Virginia, USA',
    'us-west1': 'The Dalles, Oregon, USA',
    'us-west2': 'Los Angeles, California, USA',
    'us-west3': 'Salt Lake City, Utah, USA',
    'us-west4': 'Las Vegas, Nevada, USA',
}


class GCEProvider(BaseLibcloudProvider):
    alias = 'gce'
    default_size = 'n1-standard-1'

    def __init__(self, user_id, key, project, region=None, **kwargs):
        constructor = get_driver(Provider.GCE)
        self.driver = constructor(user_id=user_id, key=key, project=project, auth_type='SA')
        self.region = region


    def get_sizes(self, **kwargs):
        sizes = []
        # TODO: Integrate with pricing API, these will be estimates based on 2020-01-31 Iowa pricing
        cpu_cost = 16.153221
        memory_cost = 2.165107
        for size in self.driver.list_sizes(self.region or 'us-east1-b'):
            sizes.append(NodeSize(
                size.name,
                size.ram/2**10,
                size.extra['guestCpus'],
                'No disk', # Extra volumes beyond the tiny root disk needs to be specified manually
                size.ram/2**10 * memory_cost + size.extra['guestCpus'] * cpu_cost,
                {},
            ))

        sizes.sort(key=lambda s: s.monthly_cost)
        return sizes


    def add_list_regions_arguments(self, parser):
        parser.add_argument('-z', '--include-zones', action='store_true')


    def get_regions(self, include_zones=False, **kwargs):
        regions = []
        for location in self.driver.ex_list_regions():
            pretty_name = region_pretty_names.get(location.name, location.name)
            if include_zones:
                for zone in location.zones:
                    regions.append(Region(zone.name, pretty_name))
            else:
                regions.append(Region(location.name, pretty_name))
        regions.sort(key=lambda r: r.id)
        return regions


    def create_remote_ssh_key(self, key_name, ssh_key, public_key):
        # Since we don't use global keys there's no item on GCE to represent the key
        return (key_name, 'root:%s hart@saltmaster' % public_key)


    def destroy_remote_ssh_key(self, remote_key):
        # The saltmaster discards the key after creation so don't have to remove it from the
        # metadata apart from cleaning up
        pass


    def add_create_minion_arguments(self, parser):
        def split_csv_keyval(clistring):
            ret = {}
            for key_value_pair in clistring.split(','):
                if not '=' in key_value_pair:
                    raise UserError('GCE requires labels to be key=value pairs')

                key, value = key_value_pair.split('=', 1)
                ret[key] = value
            return ret

        parser.add_argument('-z', '--zone', help='GCE zone to launch in')
        parser.add_argument('-l', '--labels', type=split_csv_keyval,
            help='Comma-separated key=value pairs of labels to add to the node.')
        parser.add_argument('--subnet', help='The subnet to launch in.')
        parser.add_argument('--volume-size', type=int, default=10,
            help='The size of the boot drive in GB, minimum 10')
        parser.add_argument('--volume-type', default='pd-ssd',
            help='Which volume type to use for the boot drive. Default: %(default)s',
            choices=('pd-standard', 'pd-ssd'))
        parser.add_argument('--enable-oslogin', action='store_true',
            help='By default OS Login is disabled, pass this flag to enable it for this minion')
        # To use the highest-performing disk on GCE (local ssd mounted over
        # NVMe) we need to mount multiple disks as that can't be the boot disk.
        # Decide on a CLI convention for specifying arbitrary additional disks.
        # Ref. https://cloud.google.com/compute/docs/disks/#introduction


    def create_node(self,
            minion_id,
            region,
            debian_codename,
            auth_key,
            cloud_init,
            private_networking,
            tags,
            size=None,
            **kwargs):
        if size is None:
            size = self.default_size

        desired_zone = kwargs.get('zone')
        if not desired_zone:
            raise UserError('You must specify the GCE zone to launch in')

        zone = self.driver.ex_get_zone(desired_zone)
        if not zone:
            raise UserError('Unknown zone %r' % desired_zone)

        subnet_string = kwargs.get('subnet')
        # The arm image looks like
        # debian-<debian-numeric-version>-<debian-codename>-arm64-v20231010
        # The x86-64 image looks like
        # debian-<debian-numeric-version>-<debian-codename>-v20231010
        # To avoid getting the wrong arch on the image we need to include enough
        # of the prefix to identify the image uniquely
        image = self.driver.ex_get_image('debian-%d-%s-v' % (
            DEBIAN_VERSIONS[debian_codename], debian_codename))
        volume_type = kwargs.get('volume_type')
        disk_type = self.driver.ex_get_disktype(volume_type, zone=zone)
        regional_subnets = self.driver.ex_list_subnetworks(region)
        subnet = get_selected_or_default_subnet(regional_subnets, subnet_string)
        node = self.driver.create_node(
            name=name_from_minion_id(minion_id),
            size=size,
            location=zone,
            image=None, # Specified in the disk params
            description=minion_id,
            ex_tags=tags,
            ex_network=subnet.network,
            ex_subnetwork=subnet,
            ex_metadata={
                'sshKeys': auth_key,
                'startup-script': cloud_init,
                # We assume that if you manage servers with salt you want to use salt to
                # manage ssh access, but we allow this to be overridden
                'enable-oslogin': kwargs.get('enable_oslogin', False),
            },
            ex_labels=kwargs.get('labels'),
            ex_disks_gce_struct=[{
                'autoDelete': True,
                'boot': True,
                'type': 'PERSISTENT', # The boot drive has to be persistent
                'mode': 'READ_WRITE',
                'initializeParams': {
                    'diskSizeGb': kwargs.get('volume_size'),
                    'diskType': disk_type.extra['selfLink'],
                    'sourceImage': image.extra['selfLink']
                }
            }],
        )
        return node, None


    def wait_for_init_script(self, client, extra=None):
        # Creds to https://stackoverflow.com/a/14158100 for a way to get the pid
        _, stdout, _ = client.exec_command(
            'echo $$ && exec tail -f -n +1 /var/log/syslog')
        tail_pid = int(stdout.readline())
        for line in stdout:
            print(line, end='')
            # TODO: This doesn't detect if the init script failed for some reason, need
            # to find a reliable way to do that
            if 'hart-init-complete' in line:
                stdout.channel.close()
                client.exec_command('kill %d' % tail_pid)
                break


    def destroy_node(self, node, extra=None, **kwargs):
        self.driver.destroy_node(node, ex_sync=False)


    def get_node(self, node):
        if isinstance(node, str):
            node = self.driver.ex_get_node(name_from_minion_id(node))
        else:
            node = self.driver.ex_get_node(node.name)
        return node


def name_from_minion_id(minion_id):
    '''
    Transform a minion id into a valid GCE VM name.

    GCE VM names has to be a valid DNS label, which doesn't allow dots, or
    leading numbers. Thus we transform minion ids into a compatible name which
    is preferably also somewhat legible from the cloud console. The full minion
    id is stored in the description. This also needs to be deterministic.
    '''
    sanitized_name = '-'.join(reversed(minion_id.split('.')))
    hashed_id = hashlib.sha256(minion_id.encode('utf-8')).hexdigest()
    return 'hart-%s-%s' % (sanitized_name[:47], hashed_id[:6])


def get_selected_or_default_subnet(subnets, subnet_string):
    if not subnets:
        raise UserError('No subnets available in the given region')

    if subnet_string is None and len(subnets) == 1:
        return subnets[0]
    elif subnet_string is None:
        raise UserError('More than one subnet available in %s, need to pick one of %s' % (
            subnets[0].region.name, ', '.join(s.name for s in subnets)))

    for subnet in subnets:
        if subnet.name == subnet_string:
            return subnet

    raise UserError('None of the subnets in %s match %r, should be one of %s' % (
        subnets[0].region.name, subnet_string, ', '.join(s.name for s in subnets)))



import datetime
import json
import time
import subprocess

from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider, NodeState
from libcloud.utils.py3 import httplib

from .base import NodeSize
from .libcloud import BaseLibcloudProvider
from ..constants import DEBIAN_VERSIONS
from ..exceptions import UserError


class VultrProvider(BaseLibcloudProvider):
    alias = 'vultr'
    default_size = '201'

    def __init__(self, token, **kwargs):
        constructor = get_driver(Provider.VULTR)
        self.driver = constructor(token)


    def create_node(self,
            minion_id,
            region,
            debian_codename,
            auth_key,
            cloud_init,
            private_networking,
            tags,
            size=None,
            **kwargs):
        if size is None:
            size = self.default_size

        size = self.get_size(size)
        image = self.get_image(debian_codename)
        location = self.get_location(region)
        # Vultr has replaced cloud-init with their own startup script
        # implementation. This doesn't seem to be reflected in their docs,
        # but a comment here indicates as much:
        # https://discuss.vultr.com/discussion/582/cloud-init-user-data-testing/p3
        script_id = self.create_temp_startup_script(minion_id, cloud_init)
        node_extra = {
            'script_id': script_id,
            'private_networking': private_networking,
            'debian_codename': debian_codename,
        }
        tag = None
        if len(tags) > 1:
            raise UserError('Can only set a single tag on vultr')
        elif tags:
            tag = tags[0]

        node = self.driver.create_node(minion_id, size, image, location, ex_ssh_key_ids=[
            auth_key.id
        ], ex_create_attr={
            'script_id': script_id,
            'notify_activate': False,
            'enable_private_network': 'yes' if private_networking else 'no',
            'hostname': minion_id,
            'tag': tag,
        })
        # Vultr has a race condition where if the ssh key is deleted too early,
        # ie before the node has read it on startup, it won't be available to
        # use for logging in. Thus we delay the return here until the node state
        # indicates it's ready to continue.
        start_time = time.time()
        while time.time() - start_time < 180:
            print('Waiting for node to boot (%s)' % node.state)
            time.sleep(2)
            node = self.get_node(node)
            if node.state != NodeState.PENDING:
                print('Node in state %s, continuing' % node.state)
                break
        else:
            # Can't auto-destroy since the API doesn't enable destroying nodes
            # before they are initialized
            raise ValueError('Failed to start node before timeout')

        return node, node_extra


    def create_temp_startup_script(self, minion_id, cloud_init):
        current_date = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H-%M-%S')
        params = {
            'name': 'temp-script-for-%s-%s' % (minion_id, current_date),
            'script': cloud_init,
        }

        result = self.driver.connection.post('/v1/startupscript/create', params)
        if result.status != httplib.OK:
            raise ValueError('Failed to create temp startupscript')

        script_id = result.object['SCRIPTID']
        return script_id


    def delete_startup_script(self, node_extra):
        if not node_extra['script_id']:
            # The delete might be called multiple times on error, prevent it
            # from trying to delete the script twice
            return

        script_id = node_extra['script_id']
        node_extra['script_id'] = None

        print('Start up script with id %d' % script_id)

        params = {'SCRIPTID': script_id}
        result = self.driver.connection.post('/v1/startupscript/destroy', params)
        if result.status != httplib.OK:
            print('Failed to delete temp startupscript %s' % script_id)


    def create_remote_ssh_key(self, key_name, ssh_key, public_key):
        # The vultr provider mistakenly returns a success bool instead of the key pair.
        # PR from 2017 that fixes it: https://github.com/apache/libcloud/pull/998
        # ¯\_(ツ)_/¯s
        self.driver.create_key_pair(key_name, public_key)
        key_pairs = self.driver.list_key_pairs()
        for key_pair in key_pairs:
            if key_pair.name == key_name:
                break
        else:
            raise ValueError('Failed to create ssh key pair')
        return key_pair, key_pair


    def get_image(self, debian_codename):
        for image in self.driver.list_images():
            if (image.extra['family'] == 'debian'
                    and image.extra['arch'] == 'x64'
                    and debian_codename in image.name):
                return image

        raise ValueError('Debian %s image not found' % debian_codename)


    def destroy_node(self, node, extra=None, **kwargs):
        if extra is not None:
            self.delete_startup_script(extra)

        # Vultr doesn't handle deleting nodes that haven't finished
        # initialization well, wait for the node to finish boot before
        # destroying it
        timeout = 180
        start_time = time.time()
        while True:
            time.sleep(3)
            node = self.get_node(node)
            if node.state == NodeState.RUNNING:
                self.driver.destroy_node(node)
                break
            if time.time() - start_time > timeout:
                raise ValueError('Timed out waiting to delete initializing node: %s' % node.id)


    def wait_for_init_script(self, client, node_extra):
        # If we delete the startup script any earlier it might not get onto the node and
        # boot might fail
        self.delete_startup_script(node_extra)

        if DEBIAN_VERSIONS[node_extra['debian_codename']] >= 11:
            # For bullseye and newer the base image uses the standard cloud init boot log
            # location and we don't need the custom logic here
            super().wait_for_init_script(client, node_extra)
            return

        # Creds to https://stackoverflow.com/a/14158100 for a way to get the pid
        _, stdout, stderr = client.exec_command('echo $$ && exec tail -f /var/log/firstboot.log')
        tail_pid = int(stdout.readline())
        for line in stdout:
            print(line, end='')
            if line.startswith('hart-init-complete'):
                stdout.channel.close()
                client.exec_command('kill %d' % tail_pid)
                return

        raise ValueError('Failed to complete init script: %s' % ''.join(stderr).strip())


    def get_sizes(self, **kwargs):
        sizes = []
        for size in self.driver.list_sizes():
            sizes.append(NodeSize(
                size.id,
                size.ram/2**10,
                size.extra['vcpu_count'],
                '%d GB SSD' % size.disk,
                size.price,
                {},
            ))

        sizes.sort(key=lambda s: s.monthly_cost)
        return sizes


    def post_connect(self, hart_node):
        if not hart_node.node_extra['private_networking']:
            return

        # For private networking to work for the node it needs to be attached
        # inside the node too. Thus get the private IP that was assigned from
        # the API and add it.
        response = self.driver.connection.get('/v1/server/list_ipv4?SUBID=%s' % hart_node.node.id)
        networks = response.object[hart_node.node.id]
        ip = None
        netmask = None
        for network in networks:
            if network['type'] == 'private':
                ip = network['ip']
                netmask = network['netmask']
                break

        if ip is None:
            raise ValueError("Couldn't find private network attached to server")

        interface = 'ens7'
        enable_network_interfaces_d(hart_node.minion_id)
        add_ip_to_device(hart_node.minion_id, 'private', interface, ip, netmask)
        bring_up_interface_with_label(hart_node.minion_id, interface)


def add_ip(minion_id, current_device_ip, ip, netmask, ip_kind):
    '''
    :param minion_id: The minion id
    :param current_device_ip: An IP the device holds today we can use to identify it.
    :param ip: The new ip to attach
    :param ip_kind: The kind of IP to add. 'reserved' or 'private'.
    '''
    _, next_label = get_network_device_name_for_ip(minion_id, current_device_ip)
    enable_network_interfaces_d(minion_id)
    add_ip_to_device(minion_id, ip_kind, next_label, ip, netmask)
    bring_up_interface_with_label(minion_id, next_label)


def get_network_device_name_for_ip(minion_id, current_device_ip):
    '''
    Return a tuple (device_name, next_available_label).
    '''
    interfaces = get_interfaces(minion_id)
    return get_device_and_next_label_from_interfaces(interfaces, current_device_ip)


def get_interfaces(minion_id):
    output = subprocess.check_output([
        'salt',
        minion_id,
        'network.interfaces',
        '--out=json',
    ]).decode('utf-8')
    return json.loads(output)[minion_id]


def get_device_and_next_label_from_interfaces(interfaces, current_device_ip):
    device_name = None
    for name, interface in interfaces.items():
        current_device_labels = set()
        for address in interface.get('inet', []):
            label = address.get('label')
            if label:
                current_device_labels.add(label)

            if address['address'] != current_device_ip:
                continue

            device_name = name

        if device_name:
            break

    if device_name is None:
        raise ValueError('Could not find network device with ip %s' %
            current_device_ip)

    next_label_num = 0
    next_label = '%s:%d' % (device_name, next_label_num)
    while next_label in current_device_labels:
        next_label_num += 1
        next_label = '%s:%d' % (device_name, next_label_num)

    return device_name, next_label


def enable_network_interfaces_d(minion_id):
    subprocess.run([
        'salt',
        minion_id,
        'file.replace',
        '/etc/network/interfaces',
        '^#source /etc/network/interfaces.d/\\*$',
        'source /etc/network/interfaces.d/*',
        'append_if_not_found=True',
    ], check=True)


def add_ip_to_device(minion_id, ip_kind, label, ip, netmask):
    '''
    :param ip_kind: What kind of IP this is. Either 'reserved' or 'private'.
    '''
    mtu = 1450 if ip_kind == 'private' else None
    lines = [
        'auto %s' % label,
        'iface %s inet static' % label,
        'address %s' % ip,
        'netmask %s' % netmask,
    ]
    if mtu:
        lines.append('mtu %d' % mtu)

    subprocess.run([
        'salt',
        minion_id,
        'file.write',
        '/etc/network/interfaces.d/20-hart-%s-ip' % ip_kind,
        'args=[%s]' % ', '.join("'%s'" % line for line in lines),
    ], check=True)


def bring_up_interface_with_label(minion_id, label):
    subprocess.run([
        'salt',
        minion_id,
        'cmd.run',
        'ifup %s' % label,
    ], check=True)



import base64
import datetime
import os
import sys
from collections import namedtuple

import jinja2

from .constants import DEBIAN_VERSIONS
from .exceptions import UserError


HartNode = namedtuple('HartNode', 'minion_id public_ip node provider ssh_key ssh_canary node_extra')

class TerminalColors:
    WARNING = '\033[33m'
    FAIL = '\033[91m'
    RESET = '\033[0m'


def create_token():
    return base64.urlsafe_b64encode(os.urandom(32)).rstrip(b'=').decode('utf-8')


def get_cloud_init_template(template_name='minion.sh'):
    template_directory = os.path.join(os.path.dirname(__file__), 'cloud-init')
    environment = jinja2.Environment(loader=jinja2.FileSystemLoader(template_directory))
    return environment.get_template(template_name)


def get_saltstack_repo_url(debian_codename, salt_branch):
    # As of 2023-08-18 salt hasn't released binaries for bookworm, but the
    # onedir installation should work here as well, thus using the bullseye
    # repo until they get a release out.
    if debian_codename == "bookworm":
        debian_codename = "bullseye"

    debian_version = DEBIAN_VERSIONS[debian_codename]
    return 'https://repo.saltproject.io/salt/py3/debian/%s/amd64/%s %s main' % (
        debian_version, salt_branch, debian_codename)


def build_ssh_key_name(minion_id):
    current_date = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H-%M-%S')
    return 'hart-temp-for-%s-at-%s' % (minion_id, current_date)


def remove_argument_from_parser(parser, arg):
    '''Remove an argument from the given parser.'''
    for action in parser._actions:
        if action.option_strings and arg in action.option_strings:
            parser._remove_action(action)

            for option_string in action.option_strings:
                del parser._option_string_actions[option_string]


def log_warning(message, end='\n'):
    log_to_stderr_with_color(message, TerminalColors.WARNING, end)


def log_error(message, end='\n'):
    log_to_stderr_with_color(message, TerminalColors.FAIL, end)


def log_to_stderr_with_color(message, color, end):
    out = sys.stderr
    if out.isatty():
        out.write(color)
        out.write(message)
        out.write(TerminalColors.RESET)
    else:
        out.write(message)
    if end:
        out.write(end)
    out.flush()
